
import os
from typing import Any, Dict, Optional, List, Union, Literal
from pydantic import BaseModel, ConfigDict, Field, model_validator

from orca.config_interfaces import IVariablesConfig


class ConfigModel(BaseModel):
    model_config = ConfigDict(extra='allow')

    @property
    def model_extra(self) -> Dict[str, Any]:
        model_extra = super().model_extra
        if model_extra is None:
            return {}
        return model_extra
    
class SystemOptionsConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')

    def get_deployment_stages(self) -> List[str]:
        return list(self.model_extra.keys())

class VariablesConfigModel(ConfigModel, IVariablesConfig):
    model_config = ConfigDict(extra='allow')
    
    def get_deployment_stages(self) -> List[str]:
        return list(self.model_extra.keys())
    
class ScriptConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    source: str

class ScriptBaseConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    base_dir: str = Field(alias='base-dir', default='')
    scripts: Dict[str, ScriptConfigModel] = Field(default={})

class MethodActionConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    resource: Optional[str] = None
    command: str
    inputs: List[str] = []
    outputs: List[str] | None = None

class MethodConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    script: List[str] = []
    on_init: List[str] = Field([], alias='on-init')
    actions: List[Dict[str, MethodActionConfigModel]]

class ThreadStepConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    method: str
    spawn: List[str] = []

class LabwareThreadConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    labware: str
    start: str
    end: str
    type: Literal['start', 'wrapper']
    scripts: List[str] = Field([])
    steps: List[Union[str, ThreadStepConfigModel]]


class SystemSettingsConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    name: str
    version: str = 'latest'
    description: str = ''

class LocationConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    teachpoint_name: str = Field(alias='teachpoint-name')

class LabwareConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    type: str
    static: bool = False

class ResourcePoolConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    type: str
    resources: List[str]

class ResourceConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    type: str
    com: Optional[str] = None
    ip: Optional[str] = None
    sim: Optional[bool] = None
    base_type: Optional[str] = Field(None, alias='base-type')
    plate_pad: Optional[str] = Field(None, alias='plate-pad')

class WorkflowConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    threads: Dict[str, LabwareThreadConfigModel]

class SystemConfigModel(ConfigModel):
    model_config = ConfigDict(extra='allow')
    system: SystemSettingsConfigModel
    options: SystemOptionsConfigModel = SystemOptionsConfigModel()
    labwares: Dict[str, LabwareConfigModel]
    config: VariablesConfigModel
    locations: Dict[str, LocationConfigModel] = {}
    resources: Dict[str, Union[ResourceConfigModel, ResourcePoolConfigModel]]
    methods: Dict[str, MethodConfigModel]
    workflows: Dict[str, WorkflowConfigModel]
    scripting: ScriptBaseConfigModel | None = Field(default=None)

    @model_validator(mode='before')
    def check_resource_type(cls, values: Dict[str, Any]):
        resources = values.get('resources', {})
        for key, resource in resources.items():
            if resource['type'] == 'pool':
                # Validate as ResourcePoolConfig
                resources[key] = ResourcePoolConfigModel(**resource)
            else:
                # Validate as ResourceConfig
                resources[key] = ResourceConfigModel(**resource)
        values['resources'] = resources
        return values





import re
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union
from orca.config_interfaces import ISystemOptionsConfig, ISystemConfig, ISystemSettingsConfig, ILabwareConfig, IVariablesConfig, ILocationConfig, IResourceConfig, IResourcePoolConfig, IWorkflowConfig, ILabwareThreadConfig, IThreadStepConfig, IMethodConfig, IMethodActionConfig, IScriptBaseConfig, IScriptConfig
from orca.yml_config_builder.configs import SystemOptionsConfigModel, ConfigModel, LabwareConfigModel, LabwareThreadConfigModel, LocationConfigModel, MethodActionConfigModel, MethodConfigModel, ResourceConfigModel, ResourcePoolConfigModel, ScriptBaseConfigModel, ScriptConfigModel, SystemConfigModel, SystemSettingsConfigModel, ThreadStepConfigModel, VariablesConfigModel, WorkflowConfigModel
from orca.yml_config_builder.variable_resolution import VariablesRegistry


T = TypeVar('T', bound=ConfigModel)


class DynamicBaseConfigModel(Generic[T]):
    def __init__(self, config: T, registry: VariablesRegistry) -> None:
        self._config: T = config
        self._registry = registry

    def get_dynamic(self, value: Any) -> Any:
        """
        Resolves dynamic values.
        If the value is a string referencing a dynamic variable, resolve it.
        If it's a dictionary or list, resolve all nested values recursively.
        """
        if isinstance(value, dict):
            return {key: self.get_dynamic(val) for key, val in value.items()}
        elif isinstance(value, list):
            return [self.get_dynamic(item) for item in value]
        elif isinstance(value, str):
            return self._registry.get(value)  # Resolve dynamic values
        return value

class DynamicSystemOptionsConfigModel(DynamicBaseConfigModel[SystemOptionsConfigModel], ISystemOptionsConfig):
    pass

class DynamicVariablesConfigModel(DynamicBaseConfigModel[VariablesConfigModel], IVariablesConfig):
    def get_deployment_stages(self) -> List[str]:
        return list(self._config.model_extra.keys())
class DynamicScriptingConfigModel(DynamicBaseConfigModel[ScriptConfigModel], IScriptConfig):
    @property
    def source(self) -> str:
        return self.get_dynamic(self._config.source)


class DynamicScriptBaseConfigModel(DynamicBaseConfigModel[ScriptBaseConfigModel], IScriptBaseConfig):    
    @property    
    def base_dir(self) -> str:
        return self.get_dynamic(self._config.base_dir)
    
    @property    
    def scripts(self) -> Dict[str, IScriptConfig]:
        # handles default case of empty scripts
        if self._config == {}:
            return {}
        return {key: DynamicScriptingConfigModel(value, self._registry) for key, value in self._config.scripts.items()}

class DynamicMethodActionConfigModel(DynamicBaseConfigModel[MethodActionConfigModel], IMethodActionConfig):    
    @property    
    def resource(self) -> Optional[str]:
        return self.get_dynamic(self._config.resource)
    
    @property    
    def command(self) -> str:
        return self.get_dynamic(self._config.command)
    
    @property    
    def inputs(self) -> List[str]:
        return [self.get_dynamic(x) for x in self._config.inputs]
    
    @property    
    def outputs(self) -> Optional[List[str]]:
        if self._config.outputs is None:
            return None
        return [self.get_dynamic(x) for x in self._config.outputs]
    
    @property
    def options(self) -> Dict[str, Any]:
        return {key: self.get_dynamic(value) for key, value in self._config.model_extra.items()}

class DynamicMethodConfigModel(DynamicBaseConfigModel[MethodConfigModel], IMethodConfig):    
    @property    
    def script(self) -> List[str]:
        return [self.get_dynamic(x) for x in self._config.script]
    
    @property    
    def on_init(self) -> List[str]:
        return [self.get_dynamic(x) for x in self._config.on_init]
    
    @property    
    def actions(self) -> List[Dict[str, IMethodActionConfig]]:
        actions: List[Dict[str, IMethodActionConfig]] = []
        for action in self._config.actions:
            actions.append({key: DynamicMethodActionConfigModel(value, self._registry) for key, value in action.items()})
        return actions
    
    @property
    def options(self) -> Dict[str, Any]:
        return {key: self.get_dynamic(value) for key, value in self._config.model_extra.items()}

class DynamicThreadStepConfigModel(DynamicBaseConfigModel[ThreadStepConfigModel], IThreadStepConfig):
    @property    
    def method(self) -> str:
        return self.get_dynamic(self._config.method)
    
    @property    
    def spawn(self) -> List[str]:
        return [self.get_dynamic(s) for s in self._config.spawn]

class DynamicLabwareThreadConfigModel(DynamicBaseConfigModel[LabwareThreadConfigModel], ILabwareThreadConfig):
    @property    
    def labware(self) -> str:
        return self.get_dynamic(self._config.labware)

    @property    
    def start(self) -> str:
        return self.get_dynamic(self._config.start)

    @property    
    def end(self) -> str:
        return self.get_dynamic(self._config.end)

    @property    
    def type(self) -> str:
        return self.get_dynamic(self._config.type)

    @property    
    def scripts(self) -> List[str]:
        return [self.get_dynamic(s) for s in self._config.scripts]

    @property    
    def steps(self) -> List[Union[str, IThreadStepConfig]]:
        steps: List[Union[str, IThreadStepConfig]] = []
        for step in self._config.steps:
            if isinstance(step, str):
                # ignore any step matching the parttern ${0}, ${1}, etc
                # this is handled later in the resolution of the steps
                if re.match(r"\$\{\d+\}", step):
                    steps.append(step)
                else:
                    steps.append(self.get_dynamic(step))
            else:
                steps.append(DynamicThreadStepConfigModel(step, self._registry))
        return steps

class DynamicSystemSettingsConfigModel(DynamicBaseConfigModel[SystemSettingsConfigModel], ISystemSettingsConfig):
    @property    
    def name(self) -> str:
        return self.get_dynamic(self._config.name)

    @property    
    def version(self) -> str:
        return self.get_dynamic(self._config.version)

    @property    
    def description(self) -> str:
        return self.get_dynamic(self._config.description)

class DynamicLocationConfigModel(DynamicBaseConfigModel[LocationConfigModel], ILocationConfig):
    @property    
    def teachpoint_name(self) -> str:
        return self.get_dynamic(self._config.teachpoint_name)

    @property
    def options(self) -> Dict[str, Any]:
        return {key: self.get_dynamic(value) for key, value in self._config.model_extra.items()}

class DynamicLabwareConfigModel(DynamicBaseConfigModel[LabwareConfigModel], ILabwareConfig):
    @property    
    def type(self) -> str:
        return self.get_dynamic(self._config.type)

    @property    
    def static(self) -> bool:
        return self.get_dynamic(self._config.static)
    
    @property
    def options(self) -> Dict[str, Any]:
        return {key: self.get_dynamic(value) for key, value in self._config.model_extra.items()}

class DynamicResourcePoolConfigModel(DynamicBaseConfigModel[ResourcePoolConfigModel], IResourcePoolConfig):
    @property    
    def type(self) -> str:
        return self.get_dynamic(self._config.type)

    @property    
    def resources(self) -> List[str]:
        return self.get_dynamic(self._config.resources)

class DynamicResourceConfigModel(DynamicBaseConfigModel[ResourceConfigModel], IResourceConfig):
    @property    
    def type(self) -> str:
        return self.get_dynamic(self._config.type)

    @property    
    def com(self) -> Optional[str]:
        return self.get_dynamic(self._config.com)

    @property    
    def ip(self) -> Optional[str]:
        return self.get_dynamic(self._config.ip)

    @property    
    def plate_pad(self) -> Optional[str]:
        return self.get_dynamic(self._config.plate_pad)
    
    @property
    def sim(self) -> Optional[bool]:
        return self.get_dynamic(self._config.sim)
    
    @property
    def base_type(self) -> Optional[str]:
        return self.get_dynamic(self._config.base_type)
    
    @property
    def options(self) -> Dict[str, Any]:
        return {key: self.get_dynamic(value) for key, value in self._config.model_extra.items()}

class DynamicWorkflowConfigModel(DynamicBaseConfigModel[WorkflowConfigModel], IWorkflowConfig):

    @property    
    def threads(self) -> Dict[str, ILabwareThreadConfig]:
        return {key: DynamicLabwareThreadConfigModel(value, self._registry) for key, value in self._config.threads.items()}

class DynamicSystemConfigModel(DynamicBaseConfigModel[SystemConfigModel], ISystemConfig):
    @property
    def deployment_stage(self) -> str:
        return self._stage
    @deployment_stage.setter
    def deployment_stage(self, value: str) -> None:
        self._stage = value

    def set_options(self, options: Dict[str, Any]) -> None:
        self._config.options = SystemOptionsConfigModel.model_validate(options)

    @property    
    def system(self) -> ISystemSettingsConfig:
        return DynamicSystemSettingsConfigModel(self._config.system, self._registry)

    @property    
    def labwares(self) -> Dict[str, ILabwareConfig]:
        return {key: DynamicLabwareConfigModel(value, self._registry) for key, value in self._config.labwares.items()}

    @property    
    def config(self) -> IVariablesConfig:
        return DynamicVariablesConfigModel(self._config.config, self._registry)

    @property    
    def locations(self) -> Dict[str, ILocationConfig]:
        return {key: DynamicLocationConfigModel(value, self._registry) for key, value in self._config.locations.items()}

    @property    
    def resources(self) -> Dict[str, Union[IResourceConfig, IResourcePoolConfig]]:
        resources_dict: Dict[str, Union[IResourceConfig, IResourcePoolConfig]] = {}
        for key, value in self._config.resources.items():
            if isinstance(value, ResourceConfigModel):
                resources_dict[key] = DynamicResourceConfigModel(value, self._registry) 
            elif isinstance(value, ResourcePoolConfigModel):
                resources_dict[key] = DynamicResourcePoolConfigModel(value, self._registry)
            else:
                raise ValueError(f"Invalid resource type: {type(value)}")
        return resources_dict

    @property    
    def methods(self) -> Dict[str, IMethodConfig]:
        return {key: DynamicMethodConfigModel(value, self._registry) for key, value in self._config.methods.items()}


    @property    
    def workflows(self) -> Dict[str, IWorkflowConfig]:
        return {key: DynamicWorkflowConfigModel(value, self._registry) for key, value in self._config.workflows.items()}

    @property    
    def scripting(self) -> IScriptBaseConfig | None:
        if self._config.scripting is None:
            return None
        return DynamicScriptBaseConfigModel(self._config.scripting, self._registry)

    @property
    def options(self) -> DynamicSystemOptionsConfigModel:
        return DynamicSystemOptionsConfigModel(self._config.options, self._registry)


import logging
from types import MappingProxyType
from typing import Dict, List, Optional
import asyncio
from orca.driver_management.driver_installer import \
    DriverInstaller, DriverLoader, DriverManager, DriverRegistryInfo, \
    InstalledDriverInfo, InstalledDriverRegistry, RemoteAvailableDriverRegistry
from orca.orca_core import OrcaCore
from orca.system.method_executor import MethodTemplate
from orca.system.system import WorkflowTemplate

orca_logger = logging.getLogger("orca")

class OrcaApi:

    @property
    def _orca(self) -> OrcaCore:
        if self.__orca is None:
            raise ValueError(
                "An orca configuration file has not been initialized.  Please load a configuration file."
            )
        return self.__orca

    def __init__(
        self,
        available_drivers_registry: str = "https://raw.githubusercontent.com/Cheshire-Labs/orca-extensions/refs/heads/main/drivers.json",
    ) -> None:
        try:
            self.__orca: Optional[OrcaCore] = None
            installed_registry = InstalledDriverRegistry()
            self._driver_manager = DriverManager(
                installed_registry,
                DriverLoader(), 
                DriverInstaller(installed_registry), 
                RemoteAvailableDriverRegistry(available_drivers_registry, installed_registry))
        except Exception as e:
            orca_logger.error(f"Failed to initialize OrcaApi: {e}")
            raise

    def load(self, config_filepath: str) -> None:
        try:
            self.__orca = OrcaCore(config_filepath, self._driver_manager)
        except Exception as e:
            orca_logger.error(f"Failed to load configuration file: {e}")
            raise

    def init(
        self,
        config_file: Optional[str] = None,
        resource_list: Optional[List[str]] = None,
        deployment_stage: Optional[str] = None,
    ) -> None:
        if config_file is not None:
            self.load(config_file)

        asyncio.run(self._orca.initialize(resource_list, deployment_stage))

    def get_deployment_stages(self) -> List[str]:
        config_file_model = self._orca.system_config
        variable_configs = config_file_model.config
        return list(variable_configs.model_extra.keys())

    def run_workflow(self,
                     workflow_name: str,
                     config_file: Optional[str] = None,
                     deployment_stage: Optional[str] = None,
                     ) -> str:
        if config_file is not None:
            self.load(config_file)
        instance_id = self._orca.create_workflow_instance(workflow_name, deployment_stage)
        loop = asyncio.get_running_loop()
        loop.create_task(self._orca.run_workflow(instance_id))
        return instance_id

    def run_method(
        self,
        method_name: str,
        start_map: Dict[str, str],
        end_map: Dict[str, str],
        config_file: Optional[str] = None,
        deployment_stage: Optional[str] = None,
                    ) -> str:

        if config_file is not None:
            self.load(config_file)
        method_id = self._orca.create_method_instance(method_name, start_map, end_map, deployment_stage)
        loop = asyncio.get_running_loop()
        loop.create_task(self._orca.run_method(method_id))
        return method_id

    def get_workflow_recipes(self) -> MappingProxyType[str, WorkflowTemplate]:
        return self._orca.system.get_workflow_templates()

    def get_method_recipes(self) -> MappingProxyType[str, MethodTemplate]:
        return self._orca.system.get_method_templates()

    def get_labware_recipes(self) -> List[str]:
        raise NotImplementedError

    def get_running_workflows(self) -> Dict[str, str]:
        raise NotImplementedError

    def get_running_methods(self) -> Dict[str, str]:
        raise NotImplementedError

    def get_locations(self) -> List[str]:
        return [location.name for location in self._orca.system.locations]

    def get_resources(self) -> List[str]:
        return [resource.name for resource in self._orca.system.resources]

    def get_equipments(self) -> List[str]:
        return [r.name for r in self._orca.system.equipments]

    def get_transporters(self) -> List[str]:
        return [r.name for r in self._orca.system.transporters]

    def stop(self) -> None:
        self._orca.stop()

    def get_installed_drivers_info(self) -> Dict[str, InstalledDriverInfo]:
        return self._driver_manager.get_installed_drivers_info()

    def get_available_drivers_info(self) -> Dict[str, DriverRegistryInfo]:
        return self._driver_manager.get_available_drivers_info()

    def install_driver(
        self, driver_name: str, driver_repo_url: Optional[str] = None
    ) -> None:
        self._driver_manager.install_driver(driver_name, driver_repo_url)

    def uninstall_driver(self, driver_name: str) -> None:
        self._driver_manager.uninstall_driver(driver_name)



from orca.resource_models.labware import Labware
from orca.resource_models.location import Location
from orca.resource_models.transporter_resource import TransporterEquipment


from typing import Iterator, List


class RouteStep:
    def __init__(self, labware: Labware, source: Location, target: Location, transporter: TransporterEquipment) -> None:
        self.labware: Labware = labware 
        self.source: Location = source
        self.target: Location = target
        self.transporter: TransporterEquipment = transporter
        self.target_loc_reservation_id: str | None = None

    def __str__(self) -> str:
        return f"Route Step (labware: {self.labware}): {self.source.name} -> {self.target.name} with {self.transporter.name}"

class Route:
    def __init__(self, labware: Labware) -> None:
        self._steps: List[RouteStep] = []
        self._labware: Labware = labware

    def add_step(self, source: Location, target: Location, transporter: TransporterEquipment) -> None:
        """Add a new step to the route."""
        step = RouteStep(self._labware, source, target, transporter)
        self._steps.append(step)

    def get_total_steps(self) -> int:
        """Return the total number of steps in the route."""
        return len(self._steps)

    def __iter__(self) -> Iterator[RouteStep]:
        """Allow iterating over the route steps."""
        return iter(self._steps)

    def __getitem__(self, index: int) -> RouteStep:
        """Allow direct access to a route step by its index."""
        return self._steps[index]

    def __len__(self) -> int:
        """Allow getting the length of the route (number of steps)."""
        return len(self._steps)


from abc import ABC, abstractmethod
import asyncio
import logging
from typing import Any, Callable, Dict, List, Union, cast
import uuid

from orca.resource_models.base_resource import LabwareLoadableEquipment
from orca.resource_models.labware import AnyLabwareTemplate, Labware, LabwareTemplate
from orca.resource_models.location import Location
from orca.resource_models.transporter_resource import TransporterEquipment
from orca.system.labware_registry_interfaces import ILabwareRegistry
from orca.system.reservation_manager import IReservationManager, LocationReservation
from orca.system.system_map import SystemMap
from orca.workflow_models.status_enums import ActionStatus

orca_logger = logging.getLogger("orca")
class IActionObserver:
    def action_notify(self, event: str, action: 'BaseAction') -> None:
        pass

class ActionObserver(IActionObserver):
    def __init__(self, callback: Callable[[str, 'BaseAction'], None]) -> None:
        self._callback = callback
    
    def action_notify(self, event: str, action: 'BaseAction') -> None:
        self._callback(event, action)

class BaseAction(ABC):
    def __init__(self) -> None:
        self._id: str = str(uuid.uuid4())
        self.__status: ActionStatus = ActionStatus.CREATED
        self._observers: List[IActionObserver] = []
        self._lock = asyncio.Lock()

    @property
    def id(self) -> str:
        return self._id

    @property
    def status(self) -> ActionStatus:
        return self._status
    
    @property
    def _status(self) -> ActionStatus:
        return self.__status

    @_status.setter
    def _status(self, status: ActionStatus) -> None:
        self.__status = status
        for observer in self._observers:
            observer.action_notify(self.__status.name.upper(), self)

    async def execute(self) -> None:
        async with self._lock:
            if self._status == ActionStatus.COMPLETED:
                return
            if self._status == ActionStatus.ERRORED:
                raise ValueError("Action has errored, cannot execute")
            try:
                await self._perform_action()
            except Exception as e:
                self._status = ActionStatus.ERRORED
                raise e        
            self._status = ActionStatus.COMPLETED


    @abstractmethod
    async def _perform_action(self) -> None:
        raise NotImplementedError
    
    def reset(self) -> None:
        self._status = ActionStatus.CREATED

    def add_observer(self, observer: IActionObserver) -> None:
        if observer not in self._observers:
            self._observers.append(observer)


class AssignedLabwareManager:
    def __init__(self,
                 labware_registry: ILabwareRegistry,
                 expected_input_templates: List[Union[LabwareTemplate, AnyLabwareTemplate]],
                 expected_output_templates: List[Union[LabwareTemplate, AnyLabwareTemplate]]) -> None:
        self._labware_reg = labware_registry
        self._expected_input_templates = expected_input_templates
        self._expected_inputs: Dict[LabwareTemplate | AnyLabwareTemplate, Labware | None] = {template: None for template in expected_input_templates}
        self._expected_output_templates = expected_output_templates
        self._expected_outputs: Dict[LabwareTemplate | AnyLabwareTemplate, Labware | None] = {template: None for template in expected_output_templates}

    @property
    def expected_input_templates(self) -> List[Union[LabwareTemplate, AnyLabwareTemplate]]:
        return self._expected_input_templates

    @property
    def expected_output_templates(self) -> List[Union[LabwareTemplate, AnyLabwareTemplate]]:
        return self._expected_output_templates

    @property
    def expected_inputs(self) -> List[Labware]:
        if any(input is None for input in self._expected_inputs.values()):
            raise ValueError("Not all expected inputs have been assigned")
        return [labware for labware in self._expected_inputs.values() if labware is not None]

    @property
    def expected_outputs(self) -> List[Labware]:
        if any(output is None for output in self._expected_outputs.values()):
            raise ValueError("Not all expected outputs have been assigned")
        return [labware for labware in self._expected_outputs.values() if labware is not None]
        
    def assign_input(self, template_slot: LabwareTemplate, input: Labware):
        if template_slot in self._expected_inputs.keys():
            self._expected_inputs[template_slot] = input
        elif any(input is None and isinstance(key, AnyLabwareTemplate) for key, input in self._expected_inputs.items()):
            for key in self._expected_inputs.keys():
                if isinstance(key, AnyLabwareTemplate):
                    self._expected_inputs[key] = input
                    break
        else:
            raise ValueError(f"No available slot for input {input}")
        # TODO: keeping this assignment simple for now
        self.assign_output(template_slot, input)
        
    def assign_output(self, template_slot: LabwareTemplate, output: Labware):
        if template_slot in self._expected_outputs.keys():
            self._expected_outputs[template_slot] = output
        elif any(output is None and isinstance(key, AnyLabwareTemplate) for key, output in self._expected_outputs.items()):
            for key in self._expected_outputs.keys():
                if isinstance(key, AnyLabwareTemplate):
                    self._expected_outputs[key] = output
                    break
        else:
            raise ValueError(f"No available slot for output {output}")


    def __str__(self) -> str:
        return f"Input Manager: {self._expected_inputs}"


class LocationAction(BaseAction):
    def __init__(self,
                 location: Location,
                 command: str,
                 assigned_labware_manager: AssignedLabwareManager,
                 options: Dict[str, Any] = {}) -> None:
        super().__init__()   
        if location.resource is None or not isinstance(location.resource, LabwareLoadableEquipment):
            raise ValueError(f"Location {location} does not have an {type(LabwareLoadableEquipment)} resource")
        self._location = location
        self._command: str = command
        self._options: Dict[str, Any] = options
        self._assigned_labware_manager: AssignedLabwareManager = assigned_labware_manager
        self.is_executing: asyncio.Event = asyncio.Event()
        self._reservation: LocationReservation = LocationReservation(location, None)

    @property
    def resource(self) -> LabwareLoadableEquipment:
        return cast(LabwareLoadableEquipment, self._location.resource)

    @property
    def location(self) -> Location:
        return self._location
    
    @property
    def expected_inputs(self) -> List[Labware]:
        return self._assigned_labware_manager.expected_inputs
    
    @property
    def expected_outputs(self) -> List[Labware]:
        return self._assigned_labware_manager.expected_outputs
    
    @property
    def reservation(self) -> LocationReservation:
        return self._reservation
    
    def assign_input(self, template_slot: LabwareTemplate, input: Labware):
        self._assigned_labware_manager.assign_input(template_slot, input)

    async def _perform_action(self) -> None:
        # TODO: check the correct labware is present
        if len(self.get_missing_input_labware()) > 0:
            raise ValueError("Not all expected labware is present")

        self._status = ActionStatus.PERFORMING_ACTION
        self.is_executing.set()
        # TODO: DELETE DELETE DELETE
        # for Labware in self.resource.loaded_labware:
        #    orca_logger.debug(f"LOADED LABWARE: {Labware.name}")

        # Execute the action
        if self.resource is not None:
            await self.resource.execute(self._command, self._options)

    def get_missing_input_labware(self) -> List[Labware]:
        loaded_labwares = self.resource.loaded_labware[:]
        missing_labware: List[Labware] = []

        for labware in self._assigned_labware_manager.expected_inputs:
            if labware not in loaded_labwares:
                missing_labware.append(labware)
            else:
                loaded_labwares.remove(labware)

        if len(missing_labware) > 0:
            self._status = ActionStatus.AWAITING_CO_THREADS
        return missing_labware
    
    def all_input_labware_present(self) -> bool:
        return len(self.get_missing_input_labware()) == 0
    
    def get_present_output_labware(self) -> List[Labware]:
        loaded_labwares = self.resource.loaded_labware[:]
        present_labware: List[Labware] = []

        for labware in self._assigned_labware_manager.expected_outputs:
            if labware in loaded_labwares:
                present_labware.append(labware)
                loaded_labwares.remove(labware)

        return present_labware
    
    def all_output_labware_removed(self) -> bool:
        return len(self.get_present_output_labware()) == 0
    
    def release_reservation(self) -> None:
        self._reservation.release_reservation()

    def __str__(self) -> str:
        return f"Location Action: {self.location} - {self._command}"

class MoveAction(BaseAction):
    def __init__(self,
                 labware: Labware,
                 source: Location,
                 target: Location,
                 transporter: TransporterEquipment) -> None:
        super().__init__()
        self._labware = labware
        self._source = source
        self._target = target
        self._transporter: TransporterEquipment = transporter
        self._reservation = LocationReservation(self._target, labware)
        self._release_reservation_on_place = True
        self._status = ActionStatus.AWAITING_MOVE_RESERVATION

    @property
    def source(self) -> Location:
        return self._source

    @property
    def target(self) -> Location:
        return self._target
    
    @property
    def transporter(self) -> TransporterEquipment:
        return self._transporter
    
    @property
    def labware(self) -> Labware:
        return self._labware
    
    @property
    def reservation(self) -> LocationReservation:
        return self._reservation
    
    def set_reservation(self, reservation: LocationReservation) -> None:
        self._reservation = reservation

    def set_release_reservation_on_place(self, release: bool) -> None:
        self._release_reservation_on_place = release
    
    async def _perform_action(self) -> None:
        if self._reservation is None:
            raise ValueError("Reservation must be set before performing action")
        if self._labware is None:
            raise ValueError("Labware must be set before performing action")

        if self._target.labware is not None:
            raise ValueError("Target location is occupied")

        self._status = ActionStatus.MOVING
        # move the labware
        await self._source.prepare_for_pick(self._labware)
        await self._target.prepare_for_place(self._labware)

        await self._transporter.pick(self._source)
        await self._source.notify_picked(self._labware)
        
        await self._transporter.place(self._target)
        await self._target.notify_placed(self._labware)
        # await notify_picked
        if self._release_reservation_on_place:
            self._reservation.release_reservation()


class LocationActionCollectionReservationRequest:
    def __init__(self, location_actions: List[LocationAction]) -> None:
        self._location_action_requests = location_actions
        self._reserved_action: LocationAction | None = None

    async def reserve_location(self, reservation_manager: IReservationManager, reference_point: Location, system_map: SystemMap) -> LocationAction:
        location_actions = self._get_ordered_location_actions(reference_point, system_map)
        for request in location_actions:
            request.reservation.request_reservation(reservation_manager)
            
        # await first location reservation completed
        done, pending = await asyncio.wait([asyncio.create_task(action.reservation.completed.wait()) for action in location_actions], return_when=asyncio.FIRST_COMPLETED)
        
        # cancel all pending tasks
        for task in pending:
            task.cancel()

        # get first completed route, release all completed, and cancel the rest
        for request in location_actions:
            if request.reservation.completed.is_set():
                if self._reserved_action is None:
                    self._reserved_action = request
                else:
                    reservation_manager.release_reservation(request.reservation.reserved_location.teachpoint_name)
            else:
                request.reservation.cancelled.set()

        if self._reserved_action is None:
            raise ValueError("No location reserved")
        return self._reserved_action

    def release_reservation(self) -> None:
        if self._reserved_action:
            self._reserved_action.reservation.release_reservation()
            self._reserved_action = None
        else:
            raise ValueError("No reservation to release")


    def _get_ordered_location_actions(self, reference_point: Location, system_map: SystemMap) -> List[LocationAction]:
        return sorted(self._location_action_requests, key=lambda x: system_map.get_distance(reference_point.teachpoint_name, x.location.teachpoint_name))


    def __str__(self) -> str:
        output =  f"Location Action Reservation: Resource Pool: {[r.location.teachpoint_name for r in self._location_action_requests]}"
        if self._reserved_action:
            output += f" - Reserved Location: {self._reserved_action.location.teachpoint_name}"
        else:
            output += " - Not yet reserved"
        return output



import uuid
from typing import Dict, List

from orca.system.thread_manager import IThreadManager
from orca.workflow_models.labware_thread import LabwareThread
from orca.workflow_models.workflow_templates import ThreadTemplate


class WorkflowThreadManager:
    def __init__(self, thread_manager: IThreadManager) -> None:
        self._system_thread_manager: IThreadManager = thread_manager
        self._start_threads: Dict[str, LabwareThread] = {}
        self._workflow_thread: Dict[str, LabwareThread] = {}
    
    @property
    def start_threads(self) -> List[LabwareThread]:
        return list(self._start_threads.values())
    
    @property
    def threads(self) -> List[LabwareThread]:
        return list(self._workflow_thread.values())

    def add_start_thread(self, template: ThreadTemplate) -> LabwareThread:
        thread = self.add_thread(template) 
        self._start_threads[thread.id] = thread
        return thread
    
    def add_thread(self, template: ThreadTemplate) -> LabwareThread:
        thread = self._system_thread_manager.create_thread_instance(template)
        self._workflow_thread[thread.id] = thread
        return thread
    
    async def start_all_threads(self) -> None:
        # TODO: change this to only start the start threads
        await self._system_thread_manager.start_all_threads()

class Workflow:

    def __init__(self, name:str, thread_manager: IThreadManager) -> None:
        self._id = str(uuid.uuid4())
        self._name = name
        self._thread_manager: WorkflowThreadManager = WorkflowThreadManager(thread_manager)

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def start_threads(self) -> List[LabwareThread]:
        return self._thread_manager.start_threads
    
    @property
    def threads(self) -> List[LabwareThread]:
        return self._thread_manager.threads

    def add_start_thread(self, template: ThreadTemplate) -> None:
        self._thread_manager.add_start_thread(template)

    def add_thread(self, template: ThreadTemplate) -> None:
        self._thread_manager.add_start_thread(template)

    async def start(self) -> None:
        await self._thread_manager.start_all_threads()

    def pause(self) -> None:
        raise NotImplementedError

    def resume(self) -> None:
        raise NotImplementedError
    
    def stop(self) -> None:
        raise NotImplementedError


import logging
from typing import Optional

from orca.driver_management.drivers.simulation_base.simulation_base import SimulationBaseDriver
from orca_driver_interface.driver_interfaces import ILabwarePlaceableDriver

orca_logger = logging.getLogger("orca")
class SimulationLabwarePlaceableDriver(SimulationBaseDriver, ILabwarePlaceableDriver):

    async def prepare_for_pick(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        orca_logger.info(f"Driver: {self._name} preparing for pick...")
        self._sleep()
        orca_logger.info(f"Driver: {self._name} prepared for pick")

    async def prepare_for_place(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        orca_logger.info(f"Driver: {self._name} preparing for place...")
        self._sleep()
        orca_logger.info(f"Driver: {self._name} prepared for place")

    async def notify_picked(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        orca_logger.info(f"Driver: {self._name} notified picked...")
        self._sleep()
        orca_logger.info(f"Driver: {self._name} notified picked")

    async def notify_placed(self, labware_name: str, labware_type: str, barcode: Optional[str] = None, alias: Optional[str] = None) -> None:
        orca_logger.info(f"Driver: {self._name} notified placed...")
        self._sleep()
        orca_logger.info(f"Driver: {self._name} notified placed")



from logging import Handler, LogRecord
import socketio # type: ignore
import asyncio
from typing import Dict

class SocketIOHandler(Handler):
    """A logging handler that emits records via Socket.IO."""

    def __init__(self, sio: socketio.AsyncServer):
        super().__init__()
        self.sio = sio

    def emit(self, record: LogRecord) -> None:
        """Emit a log record via Socket.IO."""

        try:
            message = {"data": self.format(record)}
            loop = asyncio.get_running_loop()
            loop.create_task(self._send_log_message(message))
        except Exception as e:
            print(f"Error sending log message: {e}")

    async def _send_log_message(self, message: Dict[str,str]) -> None:
        """Coroutine to send log message via Socket.IO."""
        try:
            await self.sio.emit("logMessage", message, namespace="/logging") # type: ignore
        except Exception as e:
            print(f"Failed to emit log message: {e}")


from typing import Any, Dict, List
from orca.yml_config_builder.configs import ConfigModel

class IVariableLookupAdapter:
    def get(self, name: str) -> Any:
        raise NotImplementedError()

class ConfigModelLookupAdapter(IVariableLookupAdapter):
    def __init__(self, config: ConfigModel) -> None:
        self._config = config

    def get(self, name: str) -> Any:
        if hasattr(self._config, name):
            value = getattr(self._config, name)
        elif hasattr(self._config, "model_extra") and hasattr(self._config.model_extra, name):
            value = self._config.model_extra[name]
        else:
            raise ValueError(f"Variable {name} not found")
        return value
    
class DictLookupAdapter(IVariableLookupAdapter):
    def __init__(self, config: Dict[str, Any]) -> None:
        self._config = config

    def get(self, name: str) -> Any:
        if name in self._config.keys():
            return self._config[name]
        raise ValueError(f"Variable {name} not found")

class VariablesLookupRouter:
    def __init__(self) -> None:
        self._registry: Dict[str, IVariableLookupAdapter] = {}

    def add_config(self, selector: str, config: ConfigModel | Dict[str, Any]) -> None:
        if isinstance(config, dict):
            self._registry[selector] = DictLookupAdapter(config)
        else:
            self._registry[selector] = ConfigModelLookupAdapter(config)

    def get_config(self, selector: str) -> IVariableLookupAdapter:
        if selector not in self._registry.keys():
            raise ValueError(f"Selector {selector} not found")
        return self._registry[selector]


class VariableGroup:
    opener = "${"
    closer = "}"
    selector_separator = ":"
    property_separator = "."
    def __init__(self, value: str) -> None:
        if value[0:2] != self.opener:
            raise ValueError(f"Invalid variable group: {value}")
        if value[-1] != self.closer:
            raise ValueError(f"Invalid variable group: {value}")
        self._value = value[2:-1]
        selector_split = self._value.split(self.selector_separator)
        self.selector: str = selector_split[0]
        self.property_names: List[str] = selector_split[1].split(self.property_separator)


class VariableStringParser:
    opener = "${"
    closer = "}"

    def __init__(self, router: VariablesLookupRouter) -> None:
        self._lookup_router = router

    def parse(self, value: List[str]) -> str:
        is_opened = False
        constant = ""
        variable_string = ""
        while len(value) > 0:
            c = value.pop(0)
            if not is_opened:
                constant += c
                if constant.endswith(self.opener):
                    constant = constant.rstrip(self.opener)
                    variable_string = self.opener 
                    is_opened = True
                    continue
            else:
                variable_string += c

                # if a another opener is found
                if len(variable_string) > 2 and variable_string.endswith(self.opener):
                    variable_string = variable_string.rstrip(self.opener)
                    prepend = list(self.opener)
                    prepend.extend(value)
                    value = prepend
                    variable_string += self.parse(value)

                if variable_string.endswith(self.closer):
                    return self._resolve_variable_value(variable_string)

        return constant

    def _resolve_variable_value(self, variable_string: str) -> Any:
        var_group = VariableGroup(variable_string)
        config: IVariableLookupAdapter = self._lookup_router.get_config(var_group.selector)
        value = config.get(var_group.property_names[0])
        if isinstance(value, dict):
            for prop_name in var_group.property_names[1:]:
                value = value[prop_name]
        # allows for recursion allowing the look up of any returned variable configurations
        if isinstance(value, str):
            return self.parse(list(value))
        return value


class VariablesRegistry:
    def __init__(self) -> None:
        self._router = VariablesLookupRouter()
        self._parser = VariableStringParser(self._router)

    def set_selector_configuration(self, selector: str, config: ConfigModel | Dict[str, Any]) -> None:
        self._router.add_config(selector, config)

    def get(self, key: Any) -> Any:
        if isinstance(key, str):
            return self._parser.parse(list(key))
        return key


from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
import logging
from typing import Callable, List
import uuid
from orca.resource_models.location import Location
from orca.resource_models.labware import AnyLabwareTemplate, Labware, LabwareTemplate
from orca.system.move_handler import MoveHandler
from orca.system.reservation_manager import IReservationManager
from orca.system.system_map import SystemMap
from orca.workflow_models.action import BaseAction, IActionObserver, LocationAction, MoveAction
from orca.workflow_models.dynamic_resource_action import DynamicResourceAction
from orca.workflow_models.status_enums import ActionStatus, MethodStatus, LabwareThreadStatus

orca_logger = logging.getLogger("orca")
class IThreadObserver(ABC):
    @abstractmethod
    def thread_notify(self, event: str, thread: LabwareThread) -> None:
        raise NotImplementedError
    
class ThreadObserver(IThreadObserver):
    def __init__(self, callback: Callable[[str, LabwareThread], None]) -> None:
        self._callback = callback
    
    def thread_notify(self, event: str, thread: LabwareThread) -> None:
        self._callback(event, thread) 

class IMethodObserver(ABC):
    @abstractmethod
    def method_notify(self, event: str, method: Method) -> None:
        raise NotImplementedError

class MethodObserver(IMethodObserver):
    def __init__(self, callback: Callable[[str, Method], None]) -> None:
        super().__init__()
        self._callback = callback
    
    def method_notify(self, event: str, method: Method) -> None:
        self._callback(event, method)
    
class Method(IActionObserver):

    def __init__(self, name: str) -> None:
        self._id = str(uuid.uuid4())
        self._name = name
        self._steps: List[DynamicResourceAction] = []
        self._current_action: LocationAction | None = None
        self.__status: MethodStatus = MethodStatus.CREATED
        self._observers: List[IMethodObserver] = []

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name
    
    @property
    def status(self) -> MethodStatus:
        return self._status

    @property
    def _status(self) -> MethodStatus:
        return self.__status

    def _set_status(self, status: MethodStatus) -> None:
        if self.__status == status:
            return
        self.__status = status
        for observer in self._observers:
            observer.method_notify(self._status.name.upper(), self)

    def assign_thread(self, input_template: LabwareTemplate, thread: LabwareThread) -> None:
        for step in self._steps:
            if input_template in step.expected_input_templates:
                step.assign_input(input_template, thread.labware)
            elif any(isinstance(template, AnyLabwareTemplate) for template in step.expected_input_templates):
                step.assign_input(input_template, thread.labware)
    
    def append_step(self, step: DynamicResourceAction) -> None:
        self._steps.append(step)
    
    @property 
    def pending_steps(self) -> List[DynamicResourceAction]:
        return [step for step in self._steps if step.status != ActionStatus.COMPLETED]

    def has_completed(self) -> bool:
        return self._status == MethodStatus.COMPLETED

    def action_notify(self, event: str, action: BaseAction) -> None:
        if event == ActionStatus.COMPLETED.name.upper():
            if len(self.pending_steps) == 0:
                self._set_status(MethodStatus.COMPLETED)
        else:
            self._set_status(MethodStatus.IN_PROGRESS)
        

    def add_observer(self, observer: IMethodObserver) -> None:
        self._observers.append(observer)

    async def resolve_next_action(self, reference_point: Location, reservation_manager: IReservationManager, system_map: SystemMap) -> LocationAction | None:
        
        if len(self.pending_steps) == 0:
            self._set_status(MethodStatus.COMPLETED)
            return None
        current_step = self.pending_steps[0]
        current_step.add_observer(self)
        return await current_step.resolve_action(reference_point, reservation_manager, system_map) 
        

class LabwareThread(IMethodObserver):

    def __init__(self, 
                 labware: Labware, 
                 start_location: Location, 
                 end_location: Location,
                 move_handler: MoveHandler,
                 reservation_manager: IReservationManager, 
                 system_map: SystemMap, 
                 observers: List[IThreadObserver] = []) -> None:
        self._labware: Labware = labware
        self._start_location: Location = start_location
        self._end_location: Location = end_location
        self._system_map: SystemMap = system_map
        self._method_sequence: List[Method] = []
        self.__status: LabwareThreadStatus = LabwareThreadStatus.UNCREATED
        self._observers = observers
        self._move_handler = move_handler
        self._reservation_manager = reservation_manager
        self._status = LabwareThreadStatus.CREATED
        # set current_location is after self._status assignment to accommodate scripts changing start location
        # TODO: source of truth needs to be changed to a labware manager
        self._current_location: Location = self._start_location 
        self._previous_action: LocationAction | None = None
        self._assigned_action: LocationAction | None = None
        self._move_action: MoveAction | None = None
        self._stop_event = asyncio.Event()
        
    @property
    def id(self) -> str:
        return self._labware.id

    @property
    def name(self) -> str:
        return self._labware.name
    
    @property
    def start_location(self) -> Location:
        return self._start_location
    
    @start_location.setter
    def start_location(self, location: Location) -> None:
        if self.status != LabwareThreadStatus.UNCREATED and self.status != LabwareThreadStatus.CREATED:
            raise ValueError("Cannot set start location.  Thread is already in progress.")
        self._start_location = location

    @property
    def end_location(self) -> Location:
        return self._end_location
    
    def set_end_location(self, location: Location) -> None:
        self._end_location = location
    
    @property
    def current_location(self) -> Location:
        return self._current_location
    
    def set_current_location(self, location: Location) -> None:
        self._current_location = location

    @property
    def status(self) -> LabwareThreadStatus:
        return self._status

    @property
    def _status(self) -> LabwareThreadStatus:
        return self.__status

    @_status.setter
    def _status(self, status: LabwareThreadStatus) -> None:
        if self.__status == status:
            return
        self.__status = status
        for observer in self._observers:
            observer.thread_notify(self._status.name.upper(), self)

    @property
    def labware(self) -> Labware:
        return self._labware
    
    @property
    def pending_methods(self) -> List[Method]:
        return [method for method in self._method_sequence if not method.has_completed()]

    def has_completed(self) -> bool:
        return self._status == LabwareThreadStatus.COMPLETED

    def initialize_labware(self) -> None:
        self._start_location.initialize_labware(self._labware)

    def append_method_sequence(self, method: Method) -> None:
        self._method_sequence.append(method)
    
    def add_observer(self, observer: IThreadObserver) -> None:
        self._observers.append(observer)

    def method_notify(self, event: str, method: Method) -> None:
        if event == MethodStatus.COMPLETED.name.upper():
            self._current_method = None

    @property
    def assigned_action(self) -> LocationAction | None:
        return self._assigned_action
        
    @property
    def move_action(self) -> MoveAction | None:
        return self._move_action

    async def start(self) -> None: 
        # initialization check
        if self.status == LabwareThreadStatus.CREATED:
            self.initialize_labware()

        while not self.has_completed():
            # no methods left, send home
            if len(self.pending_methods) == 0:
                await self._handle_thread_completion()
                return
            
            # needs next action assigned
            if self._assigned_action is None: # or self._assigned_action.status == ActionStatus.COMPLETED:
                await self._handle_action_assignment()

            assert self._assigned_action is not None

            # move to the assigned location
            while self.current_location != self._assigned_action.location and not self._stop_event.is_set():
                await self._handle_thread_move_assignment()
                await asyncio.sleep(0.2)

            if self._stop_event.is_set():
                self._handle_thread_stop()
                return
            
            await self._handle_thread_at_assigned_location()
            await asyncio.sleep(0.2)

    def stop(self) -> None:
        self._status = LabwareThreadStatus.STOPPING
        self._stop_event.set()
        
    def _handle_thread_stop(self) -> None:
        self._stop_event.clear()
        self._status = LabwareThreadStatus.STOPPED
        

    async def _handle_action_assignment(self) -> None:
        assert len(self.pending_methods) > 0
        self._status = LabwareThreadStatus.AWAITING_ACTION_RESERVATION
        current_method = self.pending_methods[0]
        # this is to fix when the lawbare is already at the location 
        # but has not released the reservation of its previous action
        current_step = current_method.pending_steps[0]
        if self._previous_action is not None and self._current_location.resource in current_step.resource_pool.resources:
            self._previous_action.release_reservation()
        
        self._assigned_action = await current_method.resolve_next_action(self.current_location, 
                                                                         self._reservation_manager, 
                                                                         self._system_map)

    async def _handle_thread_move_assignment(self) -> None:
        assert self.assigned_action is not None
        self._status = LabwareThreadStatus.AWAITING_MOVE_RESERVATION 
        self._move_action = await self._move_handler.resolve_move_action(self._labware, 
                                                                   self.current_location, 
                                                                   self.assigned_action.location, 
                                                                   self.assigned_action)
        await self._execute_move_action()

    async def _handle_thread_at_assigned_location(self) -> None:
        assert self.assigned_action is not None
        assert self.current_location == self.assigned_action.location
        # TODO: These action status should probably be asyncio events
        if self.assigned_action.all_input_labware_present():
            self._status = LabwareThreadStatus.PERFORMING_ACTION
            await self.assigned_action.execute()
        else:
            self._status = LabwareThreadStatus.AWAITING_CO_THREADS
            while self.assigned_action.status == ActionStatus.AWAITING_CO_THREADS:
                await asyncio.sleep(0.2)
            self._status = LabwareThreadStatus.PERFORMING_ACTION
        while self.assigned_action.status == ActionStatus.PERFORMING_ACTION:
            await asyncio.sleep(0.2)
        self._previous_action = self._assigned_action
        self._assigned_action = None        
            
       
    async def _handle_thread_completion(self) -> None:
        while self.current_location != self.end_location:  
            self._status = LabwareThreadStatus.AWAITING_MOVE_RESERVATION           
            self._move_action = await self._move_handler.resolve_move_action(self._labware, 
                                                                    self.current_location, 
                                                                    self.end_location, 
                                                                    None)
            await self._execute_move_action()
            await asyncio.sleep(0.2)
        self._status = LabwareThreadStatus.COMPLETED
    
    async def _execute_move_action(self) -> None:
        assert self._move_action is not None
        self._status = LabwareThreadStatus.AWAITING_MOVE_TARGET_AVAILABILITY
        while self._move_action.target.labware is not None:
            if self._move_action.reservation.deadlocked.is_set():
                await self._handle_deadlock()
            await asyncio.sleep(0.2)
        self._status = LabwareThreadStatus.MOVING
        await self._move_action.execute()
        self.set_current_location(self._move_action.target)
        # if this was the last labware to pick from the resource, then release the reservation
        if self._previous_action and self._previous_action.all_output_labware_removed():
            self._previous_action.release_reservation()
        self._previous_action = None
        self._move_action = None

    async def _handle_deadlock(self) -> None:
        assert self._move_action is not None
        orca_logger.info(f"Thread {self.name} - Deadlock detected")
        old_target = self._move_action.target
        self._move_action = await self._move_handler.handle_deadlock(self._move_action)
        orca_logger.info(f"Thread {self.name} - Deadlock resolved - reroute target {old_target.name} to {self._move_action.target.name}")



from abc import ABC, abstractmethod
from typing import Dict
from orca.system.labware_registry_interfaces import ILabwareRegistry
from orca.system.system_map import SystemMap
from orca.workflow_models.labware_thread import Method
from orca.workflow_models.workflow import Workflow
from orca.workflow_models.workflow_templates import MethodTemplate, WorkflowTemplate
from orca.system.thread_manager import IThreadManager


class WorkflowFactory:
    def __init__(self, thread_manager: IThreadManager, labware_registry: ILabwareRegistry, system_map: SystemMap) -> None:
        self._thread_manager: IThreadManager = thread_manager
        self._labware_registry: ILabwareRegistry = labware_registry
        self._system_map: SystemMap = system_map

    def create_instance(self, template: WorkflowTemplate) -> Workflow:
        workflow = Workflow(template.name, self._thread_manager)
        for thread_template in template.start_thread_templates:
            # thread = self._thread_manager.create_thread_instance(thread_template)  probably not needed
            workflow.add_start_thread(thread_template)
        return workflow


class IWorkflowRegistry(ABC):
    @abstractmethod
    def get_workflow(self, id: str) -> Workflow:
        raise NotImplementedError

    @abstractmethod
    def add_workflow(self, workflow: Workflow) -> None:
        raise NotImplementedError

    @abstractmethod
    def create_workflow_instance(self, template: WorkflowTemplate) -> Workflow:
        raise NotImplementedError

    @abstractmethod
    def get_method(self, id: str) -> Method:
        raise NotImplementedError

    @abstractmethod
    def add_method(self, method: Method) -> None:
        raise NotImplementedError

    @abstractmethod
    def create_method_instance(self, template: MethodTemplate) -> Method:
        raise NotImplementedError


class WorkflowRegistry(IWorkflowRegistry):
    def __init__(self, thread_manager: IThreadManager, labware_registry: ILabwareRegistry, system_map: SystemMap) -> None:
        self._workflow_factory = WorkflowFactory(thread_manager, labware_registry, system_map)
        self._workflows: Dict[str, Workflow] = {}
        self._methods: Dict[str, Method] = {}

    def get_workflow(self, id: str) -> Workflow:
        return self._workflows[id]
    
    def add_workflow(self, workflow: Workflow) -> None:
        if workflow.id in self._workflows.keys():
            raise KeyError(f"Workflow {workflow.id} is already defined in the system.  Each workflow must have a unique id")
        self._workflows[workflow.id] = workflow
    
    def create_workflow_instance(self, template: WorkflowTemplate) -> Workflow:
        return self._workflow_factory.create_instance(template)
    
    def get_method(self, id: str) -> Method:
        return self._methods[id]
    
    def add_method(self, method: Method) -> None:
        if method.id in self._methods.keys():
            raise KeyError(f"Method {method.id} is already defined in the system.  Each method must have a unique id")
        self._methods[method.id] = method

    def create_method_instance(self, template: MethodTemplate) -> Method:
        # TODO: May need to delete this method and rexamine the design for method access
        raise NotImplementedError
    
    def clear(self) -> None:
        self._workflows.clear()
        self._methods.clear()


from types import MappingProxyType
from orca.resource_models.base_resource import IResource
from orca.resource_models.location import Location
from orca.system.labware_registry_interfaces import ILabwareRegistry, ILabwareTemplateRegistry
from orca.resource_models.labware import Labware, LabwareTemplate
from orca.system.resource_registry import IResourceRegistry
from orca.system.system_map import ILocationRegistry
from orca.system.template_registry_interfaces import IThreadTemplateRegistry, IMethodTemplateRegistry, IWorkflowTemplateRegistry
from orca.workflow_models.labware_thread import Method
from orca.workflow_models.workflow_templates import ThreadTemplate, MethodTemplate, WorkflowTemplate


from typing import List, Dict



    
# TODO: To be implemented
# TODO: LocationRegistry should probably just implement methods from this
class LabwareLocationManager:
    def __init__(self) -> None:
        self._labware_location: Dict[str, Location | None] = {}

    def get_labware_location(self, labware_id: str) -> Location | None:
        return self._labware_location.get(labware_id)
    
    def move_labware(self, labware_id: str, location: Location) -> None:
        self._labware_location[labware_id] = location

    def add_labware(self, labware_id: str) -> None:
        self._labware_location[labware_id] = None

    def get_labware_ids_at_location(self, location: Location) -> List[str]:
        return [labware_id for labware_id, loc in self._labware_location.items() if loc == location]
    
# TODO: To be implemented
class ResourceLocationManager:
    def __init__(self, 
                 resource_reg: IResourceRegistry, 
                 location_registry: ILocationRegistry) -> None:
        self._res_reg = resource_reg
        self._loc_reg = location_registry
        self._res_to_loc: Dict[str, Location] = {}
        self._loc_to_res: Dict[str, IResource] = {}
    
    def assign_resource_to_location(self, resource_name: str, location_name: str) -> None:
        resource = self._res_reg.get_resource(resource_name)
        location = self._loc_reg.get_location(location_name)
        self._res_to_loc[resource_name] = location
        self._loc_to_res[location_name] = resource

    def get_location_by_resource(self, resource_name: str) -> Location | None:
        return self._res_to_loc.get(resource_name, None)
    
    def get_resource_by_location(self, location_name: str) -> IResource | None:
        return self._loc_to_res.get(location_name, None)
    

class LabwareRegistry(ILabwareRegistry, ILabwareTemplateRegistry):
    def __init__(self) -> None:
        self._labwares: Dict[str, Labware] = {}
        self._labware_templates: Dict[str, LabwareTemplate] = {}

    @property
    def labwares(self) -> List[Labware]:
        return list(self._labwares.values())

    def get_labware(self, name: str) -> Labware:
        return self._labwares[name]
    
    def add_labware(self, labware: Labware) -> None:
        self._labwares[labware.name] = labware

    def get_labware_template(self, name: str) -> LabwareTemplate:
        return self._labware_templates[name]

    def add_labware_template(self, labware: LabwareTemplate) -> None:
        self._labware_templates[labware.name] = labware

    def clear(self) -> None:
        self._labwares.clear()
        self._labware_templates.clear()


class TemplateRegistry(IThreadTemplateRegistry, IWorkflowTemplateRegistry, IMethodTemplateRegistry):
    def __init__(self) -> None:
        self._labware_thread_templates: Dict[str, ThreadTemplate] = {}
        self._method_templates: Dict[str, MethodTemplate] = {}
        self._workflow_templates: Dict[str, WorkflowTemplate] = {}

    def get_labware_thread_template(self, name: str) -> ThreadTemplate:
        return self._labware_thread_templates[name]

    def get_workflow_templates(self) -> MappingProxyType[str, WorkflowTemplate]:
        return MappingProxyType(self._workflow_templates)

    def get_workflow_template(self, name: str) -> WorkflowTemplate:
        return self._workflow_templates[name]
    
    def get_method_templates(self) ->  MappingProxyType[str, MethodTemplate]:
        return MappingProxyType(self._method_templates)
    
    def get_method_template(self, name: str) -> MethodTemplate:
        return self._method_templates[name]

    def add_labware_thread_template(self, thread_template: ThreadTemplate) -> None:
        name = thread_template.name
        if name in self._labware_thread_templates.keys():
            raise KeyError(f"Labware {name} is already defined in the system.  Each labware must have a unique name")
        self._labware_thread_templates[name] = thread_template

    def add_method_template(self, method: MethodTemplate) -> None:
        name = method.name
        if name in self._method_templates.keys():
            raise KeyError(f"Method {name} is already defined in the system.  Each method must have a unique name")
        self._method_templates[name] = method

    def add_workflow_template(self, workflow: WorkflowTemplate) -> None:
        name = workflow.name
        if name in self._workflow_templates.keys():
            raise KeyError(f"Workflow {name} is already defined in the system.  Each workflow must have a unique name")
        self._workflow_templates[name] = workflow

    def clear(self) -> None:
        self._labware_thread_templates.clear()
        self._method_templates.clear()
        self._workflow_templates.clear()


from orca.resource_models.labware import Labware
from orca.resource_models.location import Location
from orca.system.labware_registry_interfaces import ILabwareRegistry
from orca.system.reservation_manager import IReservationManager
from orca.system.system_map import SystemMap
from orca.workflow_models.action import LocationAction, MoveAction
from orca.workflow_models.route import Route

import asyncio
from typing import List

class MoveActionCollectionReservationRequest:
    def __init__(self, requested_move_actions: List[MoveAction]):
        # ensure all the labware in each routestep is the same
        for move_action in requested_move_actions:
            if move_action.labware != requested_move_actions[0].labware:
                raise ValueError("All labware in a route must be the same")
        self._requested_move_actions = requested_move_actions
        self._reservation: MoveAction | None = None

    async def reserve_route(self, reservation_manager: IReservationManager) -> MoveAction:
        for action in self._requested_move_actions:
            action.reservation.request_reservation(reservation_manager)
            if action.reservation.completed.is_set():
                break
        
        # await first route reservation completed
        done, pending = await asyncio.wait([asyncio.create_task(action.reservation.completed.wait()) for action in self._requested_move_actions], return_when=asyncio.FIRST_COMPLETED)

        # Find the first completed reservation and set it as the reservation to use
        for task in done:
            completed_reservation = next((action for action in self._requested_move_actions if action.reservation.completed.is_set()), None)
            if completed_reservation and not self._reservation:
                self._reservation = completed_reservation
                break
        
        for task in pending:
            task.cancel()

        # release and cancel any other reservations
        for action in self._requested_move_actions:
            if action != self._reservation and action.reservation.completed.is_set():
                action.reservation.release_reservation()
            elif not action.reservation.completed.is_set():
                action.reservation.cancelled.set()

        if self._reservation is None:
            raise ValueError("No route reserved")

        return self._reservation


    def __str__(self) -> str:
        output =  f"Route Reservation: Requested Route Steps: {self._requested_move_actions}"
        # print out each requested route step on a separate line
        for route_step in self._requested_move_actions:
            output += f"\n - {route_step}"
        if self._reservation:
            output += f"\n - Reserved Route Step: {self._reservation.reservation.reserved_location}"
        else:
            output += "\n - Not yet reserved"
        return output


class MoveHandler:
    def __init__(self,
                reservation_manager: IReservationManager,
                labware_registry: ILabwareRegistry,
                system_map: SystemMap) -> None:
        self._reservation_manager = reservation_manager
        self._labware_registry = labware_registry
        self._system_map = system_map

    async def resolve_move_action(self, labware: Labware, current_location: Location, target_location: Location, assigned_action: LocationAction | None = None) -> MoveAction:
        potential_paths = self._get_potential_paths(current_location, target_location)
        potential_moves = self._get_potential_move_actions(labware, potential_paths)
        if assigned_action is not None:
            self._assign_reservation_to_moves(potential_moves, assigned_action)
        # check for any moves using the assigned_action's reservation
        for action in potential_moves:
            if action.reservation.completed.is_set():
                return action
        route_reservation = MoveActionCollectionReservationRequest(potential_moves)
        reserved_move_action = await route_reservation.reserve_route(self._reservation_manager)
        return reserved_move_action

    async def handle_deadlock(self, move_action: MoveAction) -> MoveAction:
        # NOTE: Although move_action does not have a reservation and does not need to be released, 
        # even though it is set as completed, it is also deadlocked.  This may lead to confusion and may need to be changed
        # due to this, this handling may work better else where

        potential_paths = self._system_map.get_shortest_paths_to_deadlock_resolution(move_action.source.teachpoint_name)
        # if move_action.target is in the potential_paths, remove it
        potential_paths = [path for path in potential_paths if move_action.target.teachpoint_name not in path]
        sorted_paths = sorted(potential_paths, key=lambda path: len(path))
        potential_moves = self._get_potential_move_actions(move_action.labware, sorted_paths)
        route_reservation = MoveActionCollectionReservationRequest(potential_moves)
        reserved_move_action = await route_reservation.reserve_route(self._reservation_manager)

        return reserved_move_action

    def _get_potential_paths(self, current_location: Location, target_location: Location) -> List[List[str]]:
        if current_location == target_location:
            raise ValueError("Source and target locations are the same")
        potential_paths = self._system_map.get_all_shortest_any_paths(current_location.teachpoint_name, target_location.teachpoint_name)
        if potential_paths == []:
            raise ValueError("No routes found between source and target")
        return potential_paths

    def _get_potential_move_actions(self, labware: Labware, potential_paths: List[List[str]]) -> List[MoveAction]:
    
        potential_actions: List[MoveAction] = []
        for path in potential_paths:
            source_location = self._system_map.get_location(path[0])
            target = self._system_map.get_location(path[1])
            transporter = self._system_map.get_transporter_between(source_location.name, target.name)
            action = MoveAction(labware, source_location, target, transporter)
            potential_actions.append(action)
        return potential_actions
    
    def _assign_reservation_to_moves(self, potential_moves: List[MoveAction], assigned_action: LocationAction) -> None:
        # TODO: Fix this later - this is a temporary fix
        # use the reservation already set by the assigned action
        for move in potential_moves:
            if move.target == assigned_action.location:
                move.set_reservation(assigned_action.reservation)
                move.set_release_reservation_on_place(False)



import argparse
import cmd
import traceback
import colorama

import shlex
from typing import Dict

from orca.cli.orca_api import OrcaApi


class OrcaCmdShell(cmd.Cmd):
    intro = "Welcome to the Orca Shell.  Type help or ? to list commands.\n"
    prompt = "\033[31morca> \033[0m"
    
    def __init__(self) -> None:
        colorama.init()
        super().__init__()
        self._orca_shell: OrcaApi = OrcaApi()
        self._parsers: Dict[str, argparse.ArgumentParser] = self._create_parsers()

    def do_load(self, arg: str):
        """Load the configuration file"""
        try:
            args = self._parsers["load"].parse_args(shlex.split(arg))
            print("Loading configuration file")
            self._orca_shell.load(args.config)
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["load"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_init(self, arg: str):
        """Initialize the lab instruments"""
        try:            
            args = self._parsers["init"].parse_args(shlex.split(arg))
            print("Initializing lab instruments")
            self._orca_shell.init(args.config, args.resources)
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["init"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_run(self, arg: str):
        """Run the workflow"""
        try:
            args = self._parsers["run"].parse_args(shlex.split(arg))
            print("Running workflow")
            self._orca_shell.run_workflow(args.workflow, args.config)
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["run"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_run_method(self, arg: str):
        """Run a specific method"""
        try:
            args = self._parsers["run_method"].parse_args(shlex.split(arg))
            print("Running method")
            self._orca_shell.run_method(args.method, args.start_map, args.end_map, args.config)
        except SystemExit as e:
            
            print(f"Argument parsing error: {e}")
             
            self._parsers["run_method"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()
    
    def do_list_workflow_recipes(self, arg: str):
        """List available workflow recipes"""
        try:
            args = self._parsers["list_workflow_recipes"].parse_args(shlex.split(arg))
            print("Available Workflow Recipes:")
            for recipe in self._orca_shell.get_workflow_recipes():
                print(f"{recipe}")
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["list_workflow_recipes"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_list_method_recipes(self, arg: str):
        """List available method recipes"""
        try:
            args = self._parsers["list_method_recipes"].parse_args(shlex.split(arg))
            print("Available Method Recipes:")
            for recipe in self._orca_shell.get_method_recipes():
                print(f"{recipe}")
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["list_method_recipes"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_list_locations(self, arg: str):
        """List all locations"""
        try:
            args = self._parsers["list_locations"].parse_args(shlex.split(arg))
            print("Locations:")
            for location in self._orca_shell.get_locations():
                print(f"{location}")
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["list_locations"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_list_resources(self, arg: str):
        """List all resources"""
        try:
            args = self._parsers["list_resources"].parse_args(shlex.split(arg))
            print("Resources:")
            for resource in self._orca_shell.get_resources():
                print(f"{resource}")
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["list_resources"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_list_equipments(self, arg: str):
        """List all equipments"""
        try:
            args = self._parsers["list_equipments"].parse_args(shlex.split(arg))
            print("Equipments:")
            for equipment in self._orca_shell.get_equipments():
                print(f"{equipment}")
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["list_equipments"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_list_transporters(self, arg: str):
        """List all transporters"""
        try:
            args = self._parsers["list_transporters"].parse_args(shlex.split(arg))
            print("Transporters:")
            for transporter in self._orca_shell.get_transporters():
                print(f"{transporter}")
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["list_transporters"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_install_driver(self, arg: str):
        """Install the driver"""
        try:
            args = self._parsers["install_driver"].parse_args(shlex.split(arg))
            print("Installing driver")
            self._orca_shell.install_driver(args.name, args.url)
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["install_driver"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_uninstall_driver(self, arg: str):
        """Uninstall the driver"""
        try:
            args = self._parsers["uninstall_driver"].parse_args(shlex.split(arg))
            print("Uninstalling driver")
            self._orca_shell.uninstall_driver(args.name)
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["uninstall_driver"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_list_installed_drivers(self, arg: str):
        """List Installed drivers"""
        try:
            args = self._parsers["list_installed_drivers"].parse_args(shlex.split(arg))
            print("Installed Drivers:")
            for name in self._orca_shell.get_installed_drivers_info().items():
                print(f"{name}")
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["list_installed_drivers"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()


    def do_list_available_drivers(self, arg: str):
        """List Available drivers"""
        try:
            args = self._parsers["list_available_drivers"].parse_args(shlex.split(arg))
            print("Available Drivers:")
            for name, info in self._orca_shell.get_available_drivers_info().items():
                print(f"{name}")
        except SystemExit as e:
            print(f"Argument parsing error: {e}")
            self._parsers["list_available_drivers"].print_help()
        except Exception as e:
            print(f"Error: {e}")
            traceback.print_exc()

    def do_exit(self, arg: str):
        """Exit the shell"""
        print("Exiting Orca Shell")
        return True
    
    def do_quit(self, arg: str):
        """Exit the shell"""
        return self.do_exit(arg)

    def _create_parsers(self):
        """Create argument parsers for each command."""
        parsers = {}

        # Parser for the 'load' command
        load_parser = argparse.ArgumentParser(prog='load', description="Load the configuration file")
        load_parser.add_argument("--config", required=True, help="Configuration file")
        parsers['load'] = load_parser

        # Parser for the 'init' command
        init_parser = argparse.ArgumentParser(prog='init', description="Initialize the lab instruments")
        init_parser.add_argument("--config", help="Configuration file")
        init_parser.add_argument("--resources", help="List of resources to initialize. All others will be excluded.")
        parsers['init'] = init_parser

        # Parser for the 'run' command
        run_parser = argparse.ArgumentParser(prog='run', description="Run the workflow")
        run_parser.add_argument("--workflow", required=True, help="Workflow to be run")
        run_parser.add_argument("--stage", help="Development stage to be run")
        run_parser.add_argument("--config", help="Configuration file")
        parsers['run'] = run_parser

        # Parser for the 'run_method' command
        run_method_parser = argparse.ArgumentParser(prog='run_method', description="Run a specific method")
        run_method_parser.add_argument("--method", required=True, help="Method to be run")
        run_method_parser.add_argument("--start-map", required=True, help="JSON Dictionary of labware to start location mapping")
        run_method_parser.add_argument("--end-map", required=True, help="JSON Dictionary of labware to end location mapping")
        run_method_parser.add_argument("--stage", help="Development stage to be run")
        run_method_parser.add_argument("--config", help="Configuration file")
        parsers['run_method'] = run_method_parser

        # Parser for the 'install_driver' command
        install_driver_parser = argparse.ArgumentParser(prog='install_driver', description="Install the driver")
        install_driver_parser.add_argument("--name", required=True, help="Driver name")
        install_driver_parser.add_argument("--url", help="URL to the driver's git repo")
        parsers['install_driver'] = install_driver_parser

        # Parser for the 'uninstall_driver' command
        uninstall_driver_parser = argparse.ArgumentParser(prog='uninstall_driver', description="Uninstall the driver")
        uninstall_driver_parser.add_argument("--name", required=True, help="Driver name")
        parsers['uninstall_driver'] = uninstall_driver_parser

        # Parser for the 'list_available_drivers' command
        list_available_drivers_parser = argparse.ArgumentParser(prog='list_available_drivers', description="List available drivers")
        parsers['list_available_drivers'] = list_available_drivers_parser

        # Parser for the 'list_installed_drivers' command
        list_installed_drivers_parser = argparse.ArgumentParser(prog='list_installed_drivers', description="List installed drivers")
        parsers['list_installed_drivers'] = list_installed_drivers_parser

        # Parser for 'list_workflow_recipes' command
        list_workflow_parser = argparse.ArgumentParser(prog='list_workflow_recipes', description="List available workflow recipes")
        parsers['list_workflow_recipes'] = list_workflow_parser

        # Parser for 'list_method_recipes' command
        list_method_parser = argparse.ArgumentParser(prog='list_method_recipes', description="List available method recipes")
        parsers['list_method_recipes'] = list_method_parser

        # Parser for 'list_locations' command
        list_locations_parser = argparse.ArgumentParser(prog='list_locations', description="List all locations")
        parsers['list_locations'] = list_locations_parser

        # Parser for 'list_resources' command
        list_resources_parser = argparse.ArgumentParser(prog='list_resources', description="List all resources")
        parsers['list_resources'] = list_resources_parser

        # Parser for 'list_equipments' command
        list_equipments_parser = argparse.ArgumentParser(prog='list_equipments', description="List all equipments")
        parsers['list_equipments'] = list_equipments_parser

        # Parser for 'list_transporters' command
        list_transporters_parser = argparse.ArgumentParser(prog='list_transporters', description="List all transporters")
        parsers['list_transporters'] = list_transporters_parser

        return parsers
    
    def help_load(self):
        """Show help for the load command"""
        self._parsers['load'].print_help()
    
    def help_init(self):
        """Show help for the init command"""
        self._parsers['init'].print_help()

    def help_run(self):
        """Show help for the run command"""
        self._parsers['run'].print_help()

    def help_run_method(self):
        """Show help for the run_method command"""
        self._parsers['run_method'].print_help()

    def help_install_driver(self):
        """Show help for the install_driver command"""
        self._parsers['install_driver'].print_help()

    def help_uninstall_driver(self):
        """Show help for the uninstall_driver command"""
        self._parsers['uninstall_driver'].print_help()

    def help_list_available_drivers(self):
        """Show help for the list_drivers command"""
        self._parsers['list_available_drivers'].print_help()

    def do_help(self, arg: str):
        if arg:
            try:
                # Try to show help for the specific command
                func_name = f"help_{arg.replace('-', '_')}"
                func = getattr(self, func_name)
                func()
            except AttributeError:
                print(f"No help available for {arg}")
        else:
            print("Available commands:")
            commands = [name[3:] for name in dir(self) if name.startswith('do_')]
            for command in commands:
                print(command)
            print("\nType 'help <command>' to get help on a specific command.")
    
if __name__ == "__main__":
    OrcaCmdShell().cmdloop()


from abc import ABC, abstractmethod
from types import MappingProxyType
from typing import List

from orca.workflow_models.workflow_templates import ThreadTemplate, MethodTemplate, WorkflowTemplate


class IThreadTemplateRegistry(ABC):
    @abstractmethod
    def get_labware_thread_template(self, name: str) -> ThreadTemplate:
        raise NotImplementedError

    @abstractmethod
    def add_labware_thread_template(self, labware_thread: ThreadTemplate) -> None:
        raise NotImplementedError


class IMethodTemplateRegistry(ABC):

    @abstractmethod
    def get_method_templates(self) -> MappingProxyType[str, MethodTemplate]:
        raise NotImplementedError

    @abstractmethod
    def get_method_template(self, name: str) -> MethodTemplate:
        raise NotImplementedError

    @abstractmethod
    def add_method_template(self, method: MethodTemplate) -> None:
        raise NotImplementedError


class IWorkflowTemplateRegistry(ABC):

    @abstractmethod
    def get_workflow_templates(self) -> MappingProxyType[str, WorkflowTemplate]:
        raise NotImplementedError

    @abstractmethod
    def get_workflow_template(self, name: str) -> WorkflowTemplate:
        raise NotImplementedError

    @abstractmethod
    def add_workflow_template(self, workflow: WorkflowTemplate) -> None:
        raise NotImplementedError


from __future__ import annotations
from typing import List, Optional


class Teachpoint:
    @staticmethod
    def load_teachpoints_from_file(file_path: str) -> List[Teachpoint]:
            import xml.etree.ElementTree as ET
            positions: List[Teachpoint] = []
            tree = ET.parse(file_path)
            root = tree.getroot()
            for teachpoint in root.findall('teachpoint'):
                name = str(teachpoint.get('name'))
                shoulder = float(teachpoint.get('shoulder', 0))
                elbow = float(teachpoint.get('elbow', 0))
                wrist = float(teachpoint.get('wrist', 0))
                positions.append(Teachpoint(name, wrist, elbow, shoulder))
            return positions
    
    def __init__(self, name: str, wrist: float, elbow: float, shoulder: float):
        self._name = name
        self._wrist = wrist
        self._elbow = elbow
        self._shoulder = shoulder

    @property
    def name(self) -> str:
         return self._name
     
    @property
    def wrist(self) -> Optional[float]:
        return self._wrist
     
    @property
    def elbow(self) -> Optional[float]:
        return self._elbow

    @property
    def shoulder(self) -> Optional[float]:
        return self._shoulder

        



import logging
from typing import Any, Dict, List, Optional


from orca.driver_management.drivers.simulation_base.simulation_base import SimulationBaseDriver
from orca.helper import FilepathReconciler
from orca.resource_models.resource_extras.teachpoints import Teachpoint
from orca_driver_interface.transporter_interfaces import ITransporterDriver

orca_logger = logging.getLogger("orca")

class SimulationRoboticArmDriver(SimulationBaseDriver, ITransporterDriver):
    def __init__(self, name: str, filepath_reconciler: FilepathReconciler,  mocking_type: Optional[str] = None, sim_time: float = 0.2) -> None:
        super().__init__(name, mocking_type, sim_time)
        self._filepath_reconciler = filepath_reconciler
        self._positions: List[str] = []

    def set_init_options(self, init_options: Dict[str, Any]) -> None:
        super().set_init_options(init_options)
        self._load_taught_positions(self._init_options)

    async def initialize(self) -> None:
        self._sleep()
        self._is_initialized = True

    async def pick(self, position_name: str, labware_type: str) -> None:
        self._validate_position(position_name)
        orca_logger.info(f"Driver: {self._name} picking from {position_name}, labware type: {labware_type} picking...")
        self._sleep()
        orca_logger.info(f"Driver: {self._name} picked from {position_name}, labware type: {labware_type} picked")

    async def place(self, position_name: str, labware_type: str) -> None:
        self._validate_position(position_name)
        orca_logger.info(f"Driver: {self._name} placing to {position_name}, labware type: {labware_type} placing...")
        self._sleep()
        orca_logger.info(f"Driver: {self._name} placed to {position_name}, labware type: {labware_type} placed")

    def _validate_position(self, position_name: str) -> None:
        if position_name not in self._positions:
            raise ValueError(f"The position '{position_name}' is not taught for {self._name}")

    def get_taught_positions(self) -> List[str]:
        return self._positions

    def _load_taught_positions(self, init_options: Dict[str, Any]) -> None:
        if "positions" in init_options:
            positions_config = init_options["positions"]
            if isinstance(positions_config, list):
                self._positions = positions_config
            elif isinstance(positions_config, str):
                positions_filepath = self._filepath_reconciler.reconcile_filepath(positions_config)
                self._positions = [t.name for t in Teachpoint.load_teachpoints_from_file(positions_filepath)]
            else:
                raise ValueError(f"Positions configuration for {self._name} is not recognized.  Must be a filepath string or list of strings naming the teachpoints")